// -Tile Engine Demo (Part II)-
// Written by J. Farrell

// This program is our tile engine from article #8, with support for
// a main character and as many random NPCs as you feel like adding.
// Movement is with the arrow keys. Press Escape to end the program.

#include "Characters.h"
#include "resource.h"

// To compile this program, remember what you need to do to set up
// Visual C++ for DirectX. Make sure the DirectX SDK directories
// for libraries and headers are FIRST on their repective lists
// (under Tools menu, go to "Options..."). Then create a new project
// using this and Script1.rc as the only source files, and remember
// to link the ddraw.lib and dxguid.lib libraries to your project!

// G L O B A L S /////////////////////////////////////////////////////
// Windows
HINSTANCE hMainInstance;
HWND hMainWindow;

// DirectDraw
LPDIRECTDRAW7        lpdd7           = NULL;
LPDIRECTDRAWSURFACE7 lpddsPrimary    = NULL;
LPDIRECTDRAWSURFACE7 lpddsBack       = NULL;
LPDIRECTDRAWSURFACE7 lpddsTileset    = NULL;
LPDIRECTDRAWSURFACE7 lpddsCharacters = NULL;
LPDIRECTDRAWCLIPPER  lpddClipper     = NULL;
DDSURFACEDESC2 ddsd;

// tile engine
BYTE byMap[300][300][2];         // the actual tiles
MAPDATA mapdata;                 // the other map data
TILE tileData[NUM_TILES];
LPTILE lptile[NUM_TILES];

// characters
LPNPC lpnpc[NUM_NPCS];
PLAYER player;

// error-handling
char lpszError[100] = "An unknown error occurred.";

//////////////////////////////////////////////////////////////////////
// -WinMain()-
// Program entry point. Handles the window setup and the contains the
// program's main loop.
//////////////////////////////////////////////////////////////////////
int APIENTRY WinMain(HINSTANCE hinstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	WNDCLASSEX winclass; // our window class
	HWND hwnd;           // handle to a window
	MSG msg;             // a message variable
	int x;

	// fill out the winclass structure
	winclass.cbSize =        sizeof(WNDCLASSEX);
	winclass.style =         CS_DBLCLKS | CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	winclass.lpfnWndProc =   WindowProc;        
	winclass.cbClsExtra =    0;                  // extra class info
	winclass.cbWndExtra =    0;                  // extra window info
	winclass.hInstance =     hinstance;	         
	winclass.hIcon =         LoadIcon(NULL, IDI_WINLOGO);
	winclass.hCursor =       LoadCursor(NULL, IDC_ARROW);
	winclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winclass.lpszMenuName =  NULL;
	winclass.lpszClassName = "Tile Engine Class";  
	winclass.hIconSm =       LoadIcon(NULL, IDI_WINLOGO);
	
	// register class
	if (!RegisterClassEx(&winclass))
		return(0);
	
	// create a window
	if (!(hwnd = CreateWindowEx(NULL,            // extended style
			     "Tile Engine Class",            // class identifier
				 "Tile Engine with Characters",  // window title
				 WS_POPUP | WS_VISIBLE,          // window style
				 0, 0,							 // initial position
				 640, 480,						 // initial size
				 NULL,                           // handle to parent
				 NULL,                           // handle to menu
				 hinstance,                      // app handle
				 NULL)))                         // extra parameters
		return(0);

	// set globals
	hMainWindow = hwnd;
	hMainInstance = hinstance;

	// initialize, test for success
	if (Initialize())
	{
		// main program loop
		while (TRUE)
		{
			// get the time
			DWORD dwStart = GetTickCount();

			// check the message queue
			if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if (msg.message == WM_QUIT) // exit loop on quit
					break;
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			// check for movement on all characters
			CheckCharacters();

			// update all characters
			UpdateCharacter(&player.move);
			for (x=0; x<mapdata.nNPCCount; x++)
				UpdateCharacter(&lpnpc[x]->move);

			// advance animations
			for (x=0; x<NUM_TILES; x++)
			{
				// update each frame counter
				if (lptile[x]->nAnimSpeed > 0)
				{
					// advance frame counter
					lptile[x]->uiFrames++;

					// check if we're finished
					if (lptile[x]->uiFrames == (UINT)lptile[x]->nAnimSpeed)
					{
						lptile[x]->uiFrames = 0;
						lptile[x] = lptile[x]->lpNext;
					}
				}
			}

			// draw and show frame
			RenderMap();
			lpddsPrimary->Flip(NULL, DDFLIP_WAIT);

			// check escape key
			if (KEYSTATE(VK_ESCAPE))
				PostQuitMessage(0);

			// lock to 30 FPS (33ms per frame)
			while ((GetTickCount() - dwStart) < 33);
		}
	}
	else
		MessageBox(hMainWindow, lpszError, "Initialization Error", MB_OK | MB_ICONEXCLAMATION);
	
	// shut everything down
	Shutdown();

	// give Windows the shutdown code
	return(msg.wParam);
}

///////////////////////////////////////////////////////////////
// -WindowProc()-
// Passes all the work onto the default handler.
///////////////////////////////////////////////////////////////
LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg,
							WPARAM wparam, LPARAM lparam)
{
    // lazy man's message handler
	return(DefWindowProc(hwnd, msg, wparam, lparam));
}

///////////////////////////////////////////////////////////////
// -CheckInput()-
// This function checks the status of the arrow keys. If the
// player is moving, update the player structure accordingly.
///////////////////////////////////////////////////////////////
void CheckCharacters(void)
{
	int x;

	// first check the keyboard to see what the player is doing
	// with the main character

	// check for movement north
	if (KEYSTATE(VK_UP) && (player.move.xMove == 0) && 
		(player.move.yMove == 0) &&
		MoveOK(player.move.xTile, player.move.yTile - 1))
	{    
		player.move.yMove = -1;
		player.move.yTile--;
		player.move.yOffset += 32;
	    player.move.nFace = FACE_NORTH;
	}

	// check for movement south
	if (KEYSTATE(VK_DOWN) && (player.move.xMove == 0) && 
		(player.move.yMove == 0) &&
		MoveOK(player.move.xTile, player.move.yTile + 1))
	{    
		player.move.yMove = 1;
		player.move.yTile++;
		player.move.yOffset -= 32;
	    player.move.nFace = FACE_SOUTH;
	}

	// check for movement west
	if (KEYSTATE(VK_LEFT) && (player.move.xMove == 0) && 
		(player.move.yMove == 0) &&
		MoveOK(player.move.xTile - 1, player.move.yTile))
	{    
		player.move.xMove = -1;
		player.move.xTile--;
		player.move.xOffset += 32;
	    player.move.nFace = FACE_WEST;
	}

	// check for movement east
	if (KEYSTATE(VK_RIGHT) && (player.move.xMove == 0) && 
		(player.move.yMove == 0) &&
		MoveOK(player.move.xTile + 1, player.move.yTile))
	{    
		player.move.xMove = 1;
		player.move.xTile++;
		player.move.xOffset -= 32;
	    player.move.nFace = FACE_EAST;
	}

	// now implement random movement for NPCs
	for (x=0; x<mapdata.nNPCCount; x++)
	{
		// choose a random number
		int nDir;
		int nRand = rand();

		// move if 128 divides nRand (bits 0-6 used) and NPC is still
		if (((nRand & 0x0000001F) == 0) && (lpnpc[x]->move.xMove == 0) && (lpnpc[x]->move.yMove == 0))
		{
			// use bit 7 to determine positive (south, east) or negative (north, west) movement
			nDir = ((nRand & 0x00000080) >> 6) - 1;

			// use bit 8 to determine horizontal or vertical movement
			if ((nRand & 0x00000100) == 0)
			{
				// move vertically
				if (MoveOK(lpnpc[x]->move.xTile, lpnpc[x]->move.yTile + nDir))
				{
					// first update NPC order to keep characters sorted
					if (nDir == -1)
					{
						// move up
						int nPointer = x;
						LPNPC lpnpcTemp;

						// first check to see if this is equal to the highest NPC
						if (lpnpc[x]->move.yTile > lpnpc[0]->move.yTile)
						{
							// x is not highest, so a reordering is required.
							// step up the array until we find the next highest NPC
							while (lpnpc[nPointer - 1]->move.yTile == lpnpc[x]->move.yTile)
								nPointer--;

							// now swap NPCs at x and nPointer
							lpnpcTemp = lpnpc[nPointer];
							lpnpc[nPointer] = lpnpc[x];
							lpnpc[x] = lpnpcTemp;
							x = nPointer;
						}
						else
						{
							// swap with highest NPC
							lpnpcTemp = lpnpc[0];
							lpnpc[0] = lpnpc[x];
							lpnpc[x] = lpnpcTemp;
							x = 0;
						}
					}
					else
					{
						// move down
						int nPointer = x;
						LPNPC lpnpcTemp;

						// first check to see if this is equal to the lowest NPC
						if (lpnpc[x]->move.yTile < lpnpc[mapdata.nNPCCount - 1]->move.yTile)
						{
							// x is not lowest, so a reordering is required.
							// step down the array until we find the next lowest NPC
							while (lpnpc[nPointer + 1]->move.yTile == lpnpc[x]->move.yTile)
								nPointer++;

							// now swap NPCs at x and nPointer
							lpnpcTemp = lpnpc[nPointer];
							lpnpc[nPointer] = lpnpc[x];
							lpnpc[x] = lpnpcTemp;
							x = nPointer;
						}
						else
						{
							// swap with lowest NPC
							lpnpcTemp = lpnpc[mapdata.nNPCCount - 1];
							lpnpc[mapdata.nNPCCount - 1] = lpnpc[x];
							lpnpc[x] = lpnpcTemp;
							x = mapdata.nNPCCount - 1;
						}
					}

					// now update location info
					lpnpc[x]->move.yMove = nDir;
					lpnpc[x]->move.yTile += nDir;
					lpnpc[x]->move.yOffset -= (nDir << 5);  // remember, this is multiplication by 32
		            lpnpc[x]->move.nFace = ((nDir == 1) ? 0 : 3);
				}
			}
			else
			{
				// move horizontally
				if (MoveOK(lpnpc[x]->move.xTile + nDir, lpnpc[x]->move.yTile))
				{
					lpnpc[x]->move.xMove = nDir;
					lpnpc[x]->move.xTile += nDir;
					lpnpc[x]->move.xOffset -= (nDir << 5);
		            lpnpc[x]->move.nFace = ((nDir == 1) ? 6 : 9);
				}
			}
		}
	}
}

///////////////////////////////////////////////////////////////
// -DrawCharacter()-
// Blits a character graphic to the back buffer. The first
// argument specifies the character, and the second specifies
// which set of images should be used. Currently I am simply
// using 0 for players, 1 for NPCs. In your game when you have
// multiple player and NPC graphics, you should instead add a
// member to the NPC or player structure specifying which 
// graphic should be used.
///////////////////////////////////////////////////////////////
void DrawCharacter(LPCHARMOVE lpmove, int nIndex)
{
	// set up a default source RECT
	RECT rcSrc = {0, nIndex<<6, 32, (nIndex+1)<<6}, rcDest;
	int nFrame;

	// first figure out which frame to use
	nFrame = lpmove->nFace;
	if (((lpmove->xMove) && ((abs(lpmove->xOffset) == 4) || (abs(lpmove->xOffset) == 8))) ||
		((lpmove->yMove) && ((abs(lpmove->yOffset) == 4) || (abs(lpmove->yOffset) == 8))))
		nFrame++;
	if (((lpmove->xMove) && ((abs(lpmove->xOffset) == 20) || (abs(lpmove->xOffset) == 24))) ||
		((lpmove->yMove) && ((abs(lpmove->yOffset) == 20) || (abs(lpmove->yOffset) == 24))))
		nFrame += 2;

	// update rcSrc to correct frame
	rcSrc.left += (nFrame<<5);
	rcSrc.right += (nFrame<<5);

	// set up rcDest
	rcDest.left = (lpmove->xTile<<5) + lpmove->xOffset - mapdata.xCamera;
	rcDest.top =  (lpmove->yTile<<5) + lpmove->yOffset - mapdata.yCamera - 32;
	rcDest.right = rcDest.left + 32;
	rcDest.bottom = rcDest.top + 64;

	// blit sprite
	lpddsBack->Blt(&rcDest, lpddsCharacters, &rcSrc, DDBLT_WAIT | DDBLT_KEYSRC, NULL);
}

///////////////////////////////////////////////////////////////
// -InitCharacters()-
// Sets up all characters.
///////////////////////////////////////////////////////////////
void InitCharacters()
{
	int x;

	// What I'm going to do here is something you should *not*
	// do when creating an actual game. I'm defining characters
	// manually within the game code. This should be done with
	// an external file, such as a settings file or script. I
	// hard-code it here for two reasons: first, to show you
	// exactly what's being set and how; and second, because
	// writing all the utilities to generate scripts or settings
	// files would take too long and make this demo too
	// complicated.

	// set player location
	player.move.xTile = 18;
	player.move.yTile = 9;

	// set number of NPCs
	mapdata.nNPCCount = NUM_NPCS;

	// zero NPC structure
	for (x=0; x<mapdata.nNPCCount; x++)
		ZeroMemory(&lpnpc[x]->move, sizeof(CHARMOVE));

	// create some NPCs
	lpnpc[0]->move.xTile = 13;
	lpnpc[0]->move.yTile = 3;
	lpnpc[1]->move.xTile = 26;
	lpnpc[1]->move.yTile = 3;
	lpnpc[2]->move.xTile = 3;
	lpnpc[2]->move.yTile = 5;
	lpnpc[3]->move.xTile = 23;
	lpnpc[3]->move.yTile = 10;
	lpnpc[4]->move.xTile = 10;
	lpnpc[4]->move.yTile = 13;
	lpnpc[5]->move.xTile = 21;
	lpnpc[5]->move.yTile = 17;
}

///////////////////////////////////////////////////////////////
// -InitMap()-
// Loads the map, fills out the map data structure, and sets up
// the tile data structures. Be sure to read my comments here.
///////////////////////////////////////////////////////////////
int InitMap(void)
{
	FILE *pfile;
	int x, y;

	// first open the map
	if ((pfile = fopen("testmap.dat", "rb")) == NULL)
	{
		strcpy(lpszError, "File testmap.dat could not be opened.");
		return(FALSE);
	}

	// read the map dimensions
	mapdata.xMax = getc(pfile);
	mapdata.yMax = getc(pfile);

	// now read the actual map data and close the file
	for (x=0; x<=mapdata.xMax; x++)
	{
		for (y=0; y<=mapdata.yMax; y++)
		{
			byMap[x][y][0] = getc(pfile);
			byMap[x][y][1] = getc(pfile);
		}
	}
	fclose(pfile);

	// compute the maximum camera values
	mapdata.xMaxCamera = (mapdata.xMax - 19) * 32;
	if (mapdata.xMaxCamera < 0)
		mapdata.xMaxCamera = 0;
	mapdata.yMaxCamera = (mapdata.yMax - 14) * 32;
	if (mapdata.yMaxCamera < 0)
	    mapdata.yMaxCamera = 0;

	// set the camera at the center of the map
	mapdata.xCamera = mapdata.xMaxCamera / 2;
	mapdata.yCamera = mapdata.yMaxCamera / 2;

	// IMPORTANT - set the tile pointers initially pointing
	// at their corresponding tile structures
	for (x=0; x<NUM_TILES; x++)
		lptile[x] = &tileData[x];

	// Now we're going to set up the tile structures, but this is not normally
	// how it would be done. For an actual game tileset, which can get quite
	// large, hard-coding all this information is a very bad idea. What you'd
	// actually want to do is to create a data file that contains all of this
	// information, so the game can just open the data file and read in what
	// it needs. Because this example tileset is so small, and to keep things
	// easy to understand (i.e. not having tons of data files), I'm just going
	// to fill out the structures here.

	// initialize structures
	for (x=0; x<NUM_TILES; x++)
	{
		// set all the RECTs on the surface
		RECT rcTemp = {0, x<<5, 32, (x+1)<<5};
		tileData[x].rcLocation = rcTemp;

		// reset variables
		tileData[x].uiFrames = 0;
		tileData[x].bWalkOK = FALSE;
	}

	// set bWalkOK for grass
	tileData[1].bWalkOK = TRUE;

	// return success
	return(TRUE);
}

///////////////////////////////////////////////////////////////
// -Initialize()-
// Creates all the DirectX objects we'll need for this demo,
// sets the screen resolution to 640x480x16.
///////////////////////////////////////////////////////////////
int Initialize(void)
{
	int x;

	// create the DirectDraw7 interface
	if (FAILED(DirectDrawCreateEx(NULL, (void**)&lpdd7, IID_IDirectDraw7, NULL)))
	{
		strcpy(lpszError, "DirectDraw interface could not be created.");
		return(FALSE);
	}

	// set cooperative level to fullscreen
	if (FAILED(lpdd7->SetCooperativeLevel(hMainWindow, DDSCL_FULLSCREEN | 
		                                               DDSCL_EXCLUSIVE |
													   DDSCL_ALLOWREBOOT)))
	{
		strcpy(lpszError, "Fullscreen cooperation could not be set.");
		return(FALSE);
	}

	// hide the mouse
	ShowCursor(FALSE);

	// set 640x480x16 resolution
	if (FAILED(lpdd7->SetDisplayMode(640, 480, 16, 0, 0)))
	{
		strcpy(lpszError, "Resolution 640x480x16 could not be set.");
		return(FALSE);
	}

	// create our primary surface, with a back buffer
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
						  DDSCAPS_COMPLEX |
						  DDSCAPS_FLIP;
	ddsd.dwBackBufferCount = 1;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsPrimary, NULL)))
	{
		strcpy(lpszError, "Primary surface could not be created.");
		return(FALSE);
	}

	// get attached back buffer
	ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
	if (FAILED(lpddsPrimary->GetAttachedSurface(&ddsd.ddsCaps, &lpddsBack)))
	{
		strcpy(lpszError, "Attached back buffer could not be retrieved.");
		return(FALSE);
	}

	// create the clipper
	if (FAILED(lpdd7->CreateClipper(0, &lpddClipper, NULL)))
	{
		strcpy(lpszError, "Clipper interface could not be created.");
		return(FALSE);
	}

	// create and set the clip list
	LPRGNDATA lpClipList = (LPRGNDATA)malloc(sizeof(RGNDATAHEADER) +
										     sizeof(RECT));
	RECT rcBoundary = {0, 0, 640, 480};
	memcpy(lpClipList->Buffer, &rcBoundary, sizeof(RECT)); // set list
	lpClipList->rdh.dwSize = sizeof(RGNDATAHEADER);        // structure size
	lpClipList->rdh.iType = RDH_RECTANGLES;                // bound type
	lpClipList->rdh.nCount = 1;                            // num of bounds
	lpClipList->rdh.nRgnSize = sizeof(RECT);               // size of Buffer
	lpClipList->rdh.rcBound = rcBoundary;                  // bounding RECT
	if (FAILED(lpddClipper->SetClipList(lpClipList, 0)))
	{
		// always free the mallocs!
		free(lpClipList);
		strcpy(lpszError, "Clip list could not be attached.");
		return(FALSE);
	}

	// attach clipper to back buffer
	if (FAILED(lpddsBack->SetClipper(lpddClipper)))
	{
		free(lpClipList);
		strcpy(lpszError, "Clipper could not be attached to back buffer.");
		return(FALSE);
	}
	
	// free the memory
	free(lpClipList);

	// create the tileset surface (dimensions #defined in header file)
	// be sure to set a color key of black only
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth = TILESET_WIDTH;
	ddsd.dwHeight = TILESET_HEIGHT;
	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue  = 0;
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsTileset, NULL)))
	{
		strcpy(lpszError, "Tileset surface could not be created.");
		return(FALSE);
	}

	// create the characters surface (dimensions #defined in header file)
	// be sure to set a color key of black only
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth = CHARS_WIDTH;
	ddsd.dwHeight = CHARS_HEIGHT;
	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue  = 0;
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsCharacters, NULL)))
	{
		strcpy(lpszError, "Characters surface could not be created.");
		return(FALSE);
	}

	// load the images onto their respective surfaces
	if (!LoadBitmap(MAKEINTRESOURCE(BMP_TILESET), 0, 0, lpddsTileset))
	{
		strcpy(lpszError, "Tileset resource could not be loaded.");
		return(FALSE);
	}
	if (!LoadBitmap(MAKEINTRESOURCE(BMP_CHARS), 0, 0, lpddsCharacters))
	{
		strcpy(lpszError, "Characters image resource could not be loaded.");
		return(FALSE);
	}

	// allocate memory for the NPC array
	for (x=0; x<NUM_NPCS; x++)
	{
		if ((lpnpc[x] = (NPC*)malloc(sizeof(NPC))) == NULL)
		{
			strcpy(lpszError, "NPC memory could not be allocated.");
			return(FALSE);
		}
	}

	// now set up the map and characters
	if (!InitMap())
		return(FALSE);
	InitCharacters();

	// initialize random number generator
	srand(GetTickCount());

	// all set!
	return(TRUE);
}

//////////////////////////////////////////////////////////////////////
// -LoadBitmap()-
// Loads a bitmap resource onto the specified DirectDraw surface.
//////////////////////////////////////////////////////////////////////
int LoadBitmap(char *lpszName, int x, int y, LPDIRECTDRAWSURFACE7 lpdds)
{
	HBITMAP hbitmap = NULL;
	BITMAP bmp;
	HDC hdcImage, hdcSurface;

	// load bitmap resource
	if (!(hbitmap = (HBITMAP)LoadImage(hMainInstance, lpszName, 
		                               IMAGE_BITMAP, 0, 0, 
					     			   LR_CREATEDIBSECTION)))
		return(FALSE);

	// create device context for image
	if (!(hdcImage = CreateCompatibleDC(NULL)))
		return(FALSE);

	// select bitmap into the DC
	if (!(SelectObject(hdcImage, hbitmap)))
		return(FALSE);

	// get surface device context
	if (FAILED(lpdds->GetDC(&hdcSurface)))
		return(FALSE);

	// get image dimensions
	GetObject(hbitmap, sizeof(BITMAP), &bmp);
    int dx = bmp.bmWidth;
    int dy = bmp.bmHeight;

	// copy image to surface
	if (!(BitBlt(hdcSurface, x, y, dx, dy, hdcImage, 0, 0, SRCCOPY)))
		return(FALSE);
	
	// get rid of all that stuff we used
	lpdds->ReleaseDC(hdcSurface);
	DeleteDC(hdcImage);
	DeleteObject(hbitmap);

	return(TRUE);
}

///////////////////////////////////////////////////////////////
// -MoveOK()-
// Checks all the requirements necessary to allow movement onto
// the specified tile. Returns TRUE if movement is allowed, or
// FALSE otherwise.
///////////////////////////////////////////////////////////////
int MoveOK(int x, int y)
{
	int z;

    // first check the map boundaries
    if ((x < 0) || (y < 0) || (x > mapdata.xMax) || (y > mapdata.yMax))
        return(FALSE);    
	
	// now check if the main character is already there
    if ((x == player.move.xTile) && (y == player.move.yTile))
        return(FALSE);
    
	// check all the NPCs
    for (z=0; z<mapdata.nNPCCount; z++)
    {
        if ((x == lpnpc[z]->move.xTile) && (y == lpnpc[z]->move.yTile))
            return(FALSE);
    }

    // the tile itself is now the deciding factor, so just return that
    return(tileData[byMap[x][y][0]].bWalkOK);
}

//////////////////////////////////////////////////////////////////////
// -RenderCharacters()-
// Draws all characters on the back buffer. The first argument specifies
// the index of the first NPC to be drawn. If set to -1, no NPCs
// should be drawn. The last arguments are the tile range that is
// currently being viewed.
//////////////////////////////////////////////////////////////////////
void RenderCharacters(int nFirstNPC, int xMin, int xMax, int yMax)
{
	int nIndex = nFirstNPC;
	int bMainDrawn = FALSE;

	if (nFirstNPC == -1)
	{
		// render main character only
		DrawCharacter(&player.move, 0);
	}
	else
	{
		// expand boundaries to accommodate characters currently moving
		// towards offscreen areas, but still onscreen. We add 2 to yMax
		// because characters are two tiles tall.
		xMin--;
		xMax++;
		yMax += 2;

		// enter infinite loop -- we'll break it when finished
		while (TRUE)
		{
			// check to see if main character should be drawn
			if ((player.move.yTile <= lpnpc[nIndex]->move.yTile) && (!bMainDrawn))
			{
				bMainDrawn = TRUE;
				DrawCharacter(&player.move, 0);
			}

			// check to see if this NPC is inside y-range
			if (lpnpc[nIndex]->move.yTile >= yMax)
				break;

			// check to see if this NPC is inside x-range
			if ((lpnpc[nIndex]->move.xTile >= xMin) &&
				(lpnpc[nIndex]->move.xTile <= xMax))
				DrawCharacter(&lpnpc[nIndex]->move, 1);

			// check to see if all NPCs have been used up
			if (nIndex == (mapdata.nNPCCount - 1))
				break;
			else
				nIndex++;
		}

		// plot main character if it hasn't already been done
		if (!bMainDrawn)
			DrawCharacter(&player.move, 0);
	}
}

//////////////////////////////////////////////////////////////////////
// -RenderMap()-
// Draws the map on the back buffer.
//////////////////////////////////////////////////////////////////////
void RenderMap(void)
{
	BYTE byTile;

	// center camera on main character
	mapdata.xCamera = (player.move.xTile<<5) + player.move.xOffset - 304;
	mapdata.yCamera = (player.move.yTile<<5) + player.move.yOffset - 240;

	// clip camera to map boundaries
	if (mapdata.xCamera < 0)
		mapdata.xCamera = 0;
	if (mapdata.yCamera < 0)
		mapdata.yCamera = 0;
	if (mapdata.xCamera > mapdata.xMaxCamera)
		mapdata.xCamera = mapdata.xMaxCamera;
	if (mapdata.yCamera > mapdata.yMaxCamera)
		mapdata.yCamera = mapdata.yMaxCamera;

	// set original destination RECT for first tile -- aligned with the
	// upper-left corner of the screen
	RECT rcDest = {0, 0, 32, 32};

	// find default tile range -- divide camera coordinates by 32 and
	// use the default of 21 tiles across and 16 tiles down
	int xStart = mapdata.xCamera >> 5;
	int yStart = mapdata.yCamera >> 5;
	int xEnd = xStart + 20;
	int yEnd = yStart + 15;
	int x, y;

	// now check if the camera coordinates are divisible by 32
	x = mapdata.xCamera & 0x0000001F;
	y = mapdata.yCamera & 0x0000001F;

	if (!x)
	{
		// if xCamera is divisible by 32, use only 20 tiles across
		xEnd--;
	}
	else
	{
		// otherwise move destination RECT to the left to clip the first column of tiles
		rcDest.left-=x;
		rcDest.right-=x;
	}

	if (!y)
	{
		// if yCamera is divisible by 32, use only 15 tiles down
		yEnd--;
	}
	else
	{
		// otherwise move destination RECT up to clip the first row of tiles
		rcDest.top-=y;
		rcDest.bottom-=y;
	}

	// finally make sure we're not exceeding map limits
	if (xEnd > mapdata.xMax)
		xEnd = mapdata.xMax;
	if (yEnd > mapdata.yMax)
		yEnd = mapdata.yMax;

	// store original rcDest RECT
	RECT rcTemp;
	rcTemp = rcDest;

	// plot the first layer
	for (x=xStart; x<=xEnd; x++)
	{
		for (y=yStart; y<=yEnd; y++)
		{
			// blit the tile
			byTile = byMap[x][y][0];
			lpddsBack->Blt(&rcDest, lpddsTileset, &lptile[byTile]->rcLocation, DDBLT_WAIT, NULL);

			// advance rcDest RECT
			rcDest.bottom += 32;
			rcDest.top += 32;
		}

		// reset rcDest RECT to top of next column
		rcDest.left += 32;
		rcDest.right += 32;
		rcDest.bottom -= ((yEnd - yStart + 1) << 5);
		rcDest.top -= ((yEnd - yStart + 1) << 5);
	}

	// restore the destination RECT
	rcDest = rcTemp;

	// find the first NPC in range
	int bRenderNPCs = FALSE;
	for (x=0; x<mapdata.nNPCCount; x++)
	{
		if (lpnpc[x]->move.yTile >= (yStart - 1))
		{
			if (lpnpc[x]->move.yTile <= (yEnd + 2))
				bRenderNPCs = TRUE;
			break;
		}
	}

	// render characters
	RenderCharacters((bRenderNPCs ? x : -1), xStart, xEnd, yEnd);

	// plot the second layer
	for (x=xStart; x<=xEnd; x++)
	{
		for (y=yStart; y<=yEnd; y++)
		{
			// blit the tile if necessary
			byTile = byMap[x][y][1];
			if (byTile > 0)
				lpddsBack->Blt(&rcDest, lpddsTileset, &lptile[byTile]->rcLocation, DDBLT_WAIT | DDBLT_KEYSRC, NULL);

			// advance rcDest RECT
			rcDest.bottom += 32;
			rcDest.top += 32;
		}

		// reset rcDest RECT to top of next column
		rcDest.left += 32;
		rcDest.right += 32;
		rcDest.bottom -= ((yEnd - yStart + 1) << 5);
		rcDest.top -= ((yEnd - yStart + 1) << 5);
	}
}

///////////////////////////////////////////////////////////////
// -Shutdown()-
// Release the objects we created and the memory we allocated.
///////////////////////////////////////////////////////////////
void Shutdown(void)
{
	int x;

	// release all NPC memory
	for (x=0; x<NUM_NPCS; x++)
		free(lpnpc[x]);

	if (lpddsCharacters)
	{
		// burn the characters surface!
		lpddsCharacters->Release();
		lpddsCharacters = NULL;
	}

	if (lpddsTileset)
	{
		// a pox on the tileset surface!
		lpddsTileset->Release();
		lpddsTileset = NULL;
	}

	if (lpddClipper)
	{
		// death to the clipper!
		lpddClipper->Release();
		lpddClipper = NULL;
	}

	if (lpddsBack)
	{
		// destroy the back buffer!
		lpddsBack->Release();
		lpddsBack = NULL;
	}

	if (lpddsPrimary)
	{
		// a plague upon the primary surface!
		lpddsPrimary->Release();
		lpddsPrimary = NULL;
	}

	if (lpdd7)
	{
		// blow away DirectDraw!
		lpdd7->Release();
		lpdd7 = NULL;
	}

	// dismember the window!
	CloseWindow(hMainWindow);
}

///////////////////////////////////////////////////////////////
// -UpdateCharacter()-
// Checks to see if the given movement structure specifies that
// movement is occurring. If so, update the movement data as
// necessary, and terminate movement if it is completed.
///////////////////////////////////////////////////////////////
void UpdateCharacter(LPCHARMOVE lpmove)
{
	// first check for movement south
	if (lpmove->yMove > 0)
	{
		lpmove->yOffset += 4;
		if (lpmove->yOffset == 0)
			lpmove->yMove = 0;
	}

	// now check north
	if (lpmove->yMove < 0)
	{
		lpmove->yOffset -= 4;
		if (lpmove->yOffset == 0)
			lpmove->yMove = 0;
	}

	// check movement east
	if (lpmove->xMove > 0)
	{
		lpmove->xOffset += 4;
		if (lpmove->xOffset == 0)
			lpmove->xMove = 0;
	}

	// check movement west
	if (lpmove->xMove < 0)
	{
		lpmove->xOffset -= 4;
		if (lpmove->xOffset == 0)
			lpmove->xMove = 0;
	}
}