// -Tile Engine Demo (Part II)-
// Main header file
// Written by J. Farrell

#ifndef INCLUDE_ME_ONLY_ONCE_OR_DIE

#define INCLUDE_ME_ONLY_ONCE_OR_DIE
#define WIN32_LEAN_AND_MEAN

// INCLUDES
#include <windows.h>
#include <windowsx.h>
#include <ddraw.h>
#include <stdio.h>

// DATA TYPES
typedef struct TILE_type
{
    RECT rcLocation;    // location on DirectDraw surface
    int bWalkOK;        // can the tile be walked on?
    int nAnimSpeed;     // animation speed in frames
    DWORD dwFlags;      // approach flags
    TILE_type *lpNext;  // next tile in animation
	UINT uiFrames;      // frame counter
} TILE, FAR* LPTILE;

typedef struct CHARMOVE_type
{
    int xTile, yTile;        // character's current location
    int xOffset, yOffset;    // offset from tile location in pixels
    int xMove, yMove;        // number of tiles character is currently moving
    int nFace;               // direction character is facing; takes FACE_ constants
} CHARMOVE, FAR* LPCHARMOVE;

typedef struct PLAYER_type
{
    CHARMOVE move;
} PLAYER, FAR* LPPLAYER;

typedef struct NPC_type
{
    CHARMOVE move;
} NPC, FAR* LPNPC;

// Since we're not dealing with scripts just yet, I've
// commented those fields out of the map data structure.

typedef struct MAPDATA_type
{
    int xMax, yMax;              // map size in tiles
    int xCamera, yCamera;        // camera location in pixels
    int xMaxCamera, yMaxCamera;  // max camera coordinates
    int nNPCCount;               // number of NPCs
    //int nScriptCount;            // number of linked scripts
    LPNPC lpnpc;                 // linked NPCs
    //LPSCRIPT lpscr;              // linked scripts
} MAPDATA, FAR* LPMAPDATA;

// CONSTANTS
#define NUM_NPCS          6
#define NUM_TILES        12

#define CHARS_WIDTH     384
#define CHARS_HEIGHT    128
#define TILESET_WIDTH    32
#define TILESET_HEIGHT  384

#define FACE_SOUTH 0
#define FACE_NORTH 3
#define FACE_EAST  6
#define FACE_WEST  9

// MACROS
#define INIT_DXSTRUCT(dxstruct) { ZeroMemory(&dxstruct, sizeof(dxstruct)); dxstruct.dwSize = sizeof(dxstruct); }
#define KEYSTATE(key) ((GetAsyncKeyState(key) & 0x8000) ? TRUE : FALSE)

// PROTOTYPES
LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
void CheckCharacters(void);
void DrawCharacter(LPCHARMOVE lpmove, int nIndex);
void InitCharacters(void);
int InitMap(void);
int Initialize(void);
int LoadBitmap(char *lpszName, int x, int y, LPDIRECTDRAWSURFACE7 lpdds);
int MoveOK(int x, int y);
void RenderCharacters(int nFirstNPC, int xMin, int xMax, int yMax);
void RenderMap(void);
void Shutdown(void);
void UpdateCharacter(LPCHARMOVE lpmove);

#endif